@@ -9,12 +9,14 @@ import xml.etree.ElementTree
 import numpy as np
 from typing import List, Union, Tuple
 from aston.tracefile.agilent_uv import AgilentCSDAD2
+from aston.tracefile.agilent_ms import AgilentMS
 from unithandler.base import UnitFloat
 
 # todo general
 #   - look through acaml files to see if anything else is worth parsing
 #   - work out the difference between AgilentCSDAD 1 and 2 and make DADSpectrum class general
 #   - dig for time unit in acaml file
+#   - convert to subpackage
 
 # regex for float
 _float_pattern = '\d+(.\d+)?'
@@ -72,6 +74,21 @@ def strptime_agilent_dt(dt_string: str) -> datetime.datetime:
     raise ValueError(f'The provided datetime string "{dt_string}" does not match any expected formats')
 
 
+def bisect_slice(array, minimum_value: float, maximum_value: float) -> Tuple[int, int]:
+    """
+    Finds the slice indicies for a minimum and maximum value in an array.
+
+    :param array: array like bisectable (assumes sorted)
+    :param minimum_value: minimum value
+    :param maximum_value: maximum value
+    :return: slice indicies
+    """
+    return (
+        bisect.bisect_left(array, minimum_value),
+        bisect.bisect_left(array, maximum_value) + 1,
+    )
+
+
 class DADSignalInfo(object):
     # default unit for wavelength values
     DEFAULT_WAVELENGTH_UNIT = 'nm'
@@ -400,7 +417,7 @@ class DADSpectrum(AgilentCSDAD2):
         #   - also close the open file
         self.chromatogram = self.data
 
-        # sort the data to enable indexing
+        # sort the data to enable indexing (data is stored from some middle value to max, then min to that middle value)
         max_index = self.chromatogram.columns.index(max(self.chromatogram.columns))
         reorder = [val for val in range(max_index + 1, len(self.chromatogram.columns))]
         reorder.extend([val for val in range(0, max_index + 1)])
@@ -420,6 +437,7 @@ class DADSpectrum(AgilentCSDAD2):
     @property
     def wavelengths(self) -> list:
         """list of wavelengths for the DAD"""
+        # todo check return type
         return self.chromatogram.columns
 
     @property
@@ -444,17 +462,22 @@ class DADSpectrum(AgilentCSDAD2):
         :return: tuple of indicies which encompass the band
         """
         # todo catch and raise errors if invalid indicies are generated
-        return (
-            bisect.bisect_left(
-                self.chromatogram.columns,
-                wavelength - bandwidth / 2,
-            ),
-            bisect.bisect_left(
-                self.chromatogram.columns,
-                wavelength + bandwidth / 2,
-            ) + 1  # todo set up index tests to verify slicing operates as intended
+        # todo set up index tests to verify slicing operates as intended
+        return bisect_slice(
+            self.chromatogram.columns,
+            wavelength - bandwidth / 2,
+            wavelength + bandwidth / 2,
         )
 
+    def _get_retention_indicies(self, start_time: float, end_time: float) -> Tuple[int, int]:
+        """
+        Gets the appropriate indicies to include the provided retention times in the retention time array
+
+        :param start_time: start retention time (min)
+        :param end_time: end retention time (min)
+        """
+        return bisect_slice(self.retention_times, start_time, end_time)
+
     def get_band_wavelengths(self, wavelength: float, bandwidth: float = 1.) -> list:
         """
         Returns a list of wavelengths corresponding to the band specified.
@@ -508,6 +531,20 @@ class DADSpectrum(AgilentCSDAD2):
             signal.bandwidth.real,
         )
 
+    def get_component_spectrum(self, retention_start: float, retention_end: float) -> np.ndarray:
+        """
+        Retrieves the component spectrum for the provided retention time slice.
+
+        :param retention_start: retention time start
+        :param retention_end: retention time end
+        :return:
+        """
+        start_ind, end_ind = self._get_retention_indicies(retention_start, retention_end)
+        return np.stack((
+            self.wavelengths,
+            self.chromatogram.values[start_ind:end_ind].sum(axis=0),
+        ))
+
     @classmethod
     def create_from_D_file(cls, file_path: Union[pathlib.Path, str]) -> "DADSpectrum":
         """
@@ -522,15 +559,225 @@ class DADSpectrum(AgilentCSDAD2):
             raise ValueError(f'The file path "{file_path}" does not appear to be a valid Agilent sample directory')
         uv_files = [filename for filename in os.listdir(file_path) if filename.endswith('.UV')]
         # todo catch if multiple UV files (does this happen?)
-        # todo raise error if no UV file
-        return cls(
-            os.path.join(file_path, uv_files[0])
-        )
+        try:
+            return cls(
+                os.path.join(file_path, uv_files[0])
+            )
+        except IndexError:
+            raise FileNotFoundError(f'no .UV files were found in the target path')
 
     def write_to_allotrope(self, filename: str):  # todo
         raise NotImplementedError('there is no way to do this currently')
 
 
+class MSSpectrum(AgilentMS):
+    def __init__(self,
+                 filename=None,
+                 ftype=None,
+                 data=None,
+                 ):
+        """
+        An object describing an Agilent DAD spectrum for a sample. Inherits Aston AgilentCSDAD2 and has additional
+        methods for retrieving band information.
+
+        :param filename: target filetype
+        :param ftype:
+        :param data:
+        """
+        # initialize object
+        AgilentMS.__init__(self, filename=filename, ftype=ftype, data=data)
+        # todo rewrite data attribute so that the data isn't extracted every time .data is accessed
+        self.chromatogram = self.data
+
+        # generate sorting array
+        sort_array = np.arange(self.chromatogram.columns.shape[0])[self.chromatogram.columns.argsort()]
+        self.chromatogram.columns: np.ndarray = self.chromatogram.columns[sort_array]  # sort mass array
+        # todo type hint for csr matrix
+        self.chromatogram.values = self.chromatogram.values[:, sort_array]  # rearrange csr matrix for sorted mzs
+
+    def __str__(self):
+        return f'{self.__class__.__name__} {self.info["name"]} {self.info["filename"].split(os.sep)[-1]}'
+
+    @property
+    def retention_times(self) -> np.ndarray:
+        """retention times corresponding to the data array (min)"""
+        return self.chromatogram.index
+
+    @property
+    def masses(self) -> np.ndarray:
+        """array of wavelengths for the DAD"""
+        return self.chromatogram.columns
+
+    @property
+    def summed_intensity_array(self) -> np.ndarray:
+        """returns the summed intensity array of the spectrum"""
+        return self.chromatogram.values.toarray().sum(axis=0)
+
+    @property
+    def summed_spectrum(self) -> np.ndarray:
+        """returns the mz and summed intensity array for the entire run"""
+        return np.stack((self.masses, self.summed_intensity_array))
+
+    def _get_mass_indicies(self, start_mz: float, end_mz: float = None) -> Tuple[int, int]:
+        """
+        Determines and returns the appropriate indicies for the provided masses in the mass array.
+
+        :param start_mz: start m/z ratio for the region
+        :param end_mz: end m/z ratio for the region. If not specified, the closest value to the start m/z will be
+            returned
+        """
+        return (
+            bisect.bisect_left(self.masses, start_mz),
+            bisect.bisect_left(self.masses, end_mz or start_mz) + 1,
+        )
+
+    def _get_retention_indicies(self, start_time: float, end_time: float) -> Tuple[int, int]:
+        """
+        Gets the appropriate indicies to include the provided retention times in the retention time array
+
+        :param start_time: start retention time (min)
+        :param end_time: end retention time (min)
+        """
+        return bisect_slice(self.retention_times, start_time, end_time)
+
+    def get_ion_intensities(self, start_mz: float, end_mz: float = None) -> np.ndarray:
+        """
+        Returns the intensity integral array (reconstructed single ion monitoring) for the provided ion m/z window.
+
+        :param start_mz: start m/z ratio for the region
+        :param end_mz: end m/z ratio for the region.
+        """
+        start_ind, end_ind = self._get_mass_indicies(start_mz, end_mz)
+        return self.chromatogram.values.toarray()[:, start_ind:end_ind].sum(axis=1)
+
+    def get_spectrum_of_retention_period(self, start_time: float, end_time: float) -> np.ndarray:
+        """
+        Returns the intensity array for the mass spectrum in the retention time region provided.
+
+        :param start_time: start retention time (min)
+        :param end_time: end retention time (min)
+        """
+        start_ind, end_ind = self._get_retention_indicies(start_time, end_time)
+        return self.chromatogram.values.toarray()[start_ind:end_ind].sum(axis=0)
+
+    @classmethod
+    def create_from_D_file(cls, file_path: Union[pathlib.Path, str]) -> "MSSpectrum":
+        """
+        Creates a MSSpectrum instance from an Agilent .D file
+
+        :param file_path: path to .D file
+        :return: instance
+        """
+        if type(file_path) is str:
+            file_path = pathlib.Path(file_path)
+        if file_path.suffix != '.D' or os.path.isdir(file_path) is False:
+            raise ValueError(f'The file path "{file_path}" does not appear to be a valid Agilent sample directory')
+        ms_files = [filename for filename in os.listdir(file_path) if filename.endswith('.MS')]
+        # todo catch if multiple MS files (does this happen?)
+        try:
+            return cls(
+                os.path.join(file_path, ms_files[0])
+            )
+        except IndexError:
+            raise FileNotFoundError(f'no .MS files were found in the target path')
+
+    @staticmethod
+    def _find_some_peaks(y, n_peaks: int):
+        """roughly locates 4 peaks by maximum values in the spectrum and returns their index"""
+        split = int(len(y) / n_peaks)
+        start = 0
+        end = start + split
+        splity = []
+        for i in range(n_peaks):
+            splity.append(np.asarray(y[start:end]))
+            start += split
+            end += split
+        out = []
+        for ind, section in enumerate(splity):
+            maxy = max(section)
+            if maxy == max(section[1:-1]):  # if max is not at the edge of the spectrum
+                out.append(np.where(section == maxy)[0][0] + split * ind)
+        return out
+
+    @staticmethod
+    def _calculate_resolution(x: np.ndarray, y: np.ndarray, index=None, threshold=5):
+        """
+        Finds the resolution and full width at half max of a spectrum. From PythoMS.
+
+        :param x: list of mz values
+        :param y: corresponding list of intensity values
+        :param index: index of maximum intensity (optional; used if the resolution of a specific peak is desired)
+        :param threshold: signal to noise threshold required to output a resolution
+        :return: resolution
+        """
+        y = np.asarray(y)  # convert to array for efficiency
+        if index is None:  # find index and value of maximum
+            maxy = max(y)
+            index = np.where(y == maxy)[0][0]
+        else:
+            maxy = y[index]
+        # if intensity to average is below this threshold (rough estimate of signal/noise)
+        if maxy / (sum(y) / len(y)) < threshold:
+            return None
+        halfmax = maxy / 2
+        indleft = int(index) - 1  # generate index counters for left and right walking
+        indright = int(index) + 1
+        while y[indleft] > halfmax:  # while intensity is still above halfmax
+            indleft -= 1
+        while y[indright] > halfmax:
+            indright += 1
+        return x[index] / (x[indright] - x[indleft])  # return resolution (mz over full width at half max)
+
+    def auto_resolution(self, npeaks: int = 4) -> float:
+        """
+        Attempts to automatically determine the resolution of the spectrum.
+
+        :param npeaks: number of peakds to try to find
+        :return: estimated resolution
+        """
+        intensity_array = self.summed_intensity_array
+        peak_indicies = self._find_some_peaks(
+            intensity_array,
+            n_peaks=npeaks,
+        )
+        resolutions = [
+            self._calculate_resolution(
+                self.masses,
+                intensity_array,
+                ind,
+                threshold=10,
+            )
+            for ind in peak_indicies
+        ]
+        resolutions = [res for res in resolutions if res is not None]
+        return sum(resolutions) / len(resolutions)
+
+    def extract_function_time_tic(self):
+        """duck-type method for PythoMS"""
+        pass
+
+    def get_tic_of_function(self, function: int) -> np.ndarray:
+        """duck-type method for retrieving the TIC (expected in PythoMS)"""
+        return self.chromatogram.values.toarray().sum(axis=1)
+
+    def get_timepoints_of_function(self, function: int) -> np.ndarray:
+        """duck-type method for retrieving the timepoints (expected in PythoMS)"""
+        return self.retention_times
+
+    @property
+    def functions(self):
+        """duck type function information (expected in PythoMS)"""
+        return {
+            1: {
+                'sr': [1, self.retention_times.shape[0]],
+                'nscans': self.retention_times.shape[0],
+                'type': 'MS',
+                'mode': '',
+                'window': [self.masses.argmin(), self.masses.argmax()]
+            }
+        }
+
+
 class DADSignal(DADSignalInfo):
     def __init__(self,
                  wavelength: Union[float, UnitFloat],
@@ -733,7 +980,8 @@ class HPLCSample(HPLCSampleInfo):
                  datetimestamp: Union[str, datetime.datetime],
                  method_name: str,
                  signals: Union[List[DADSignalInfo], List[DADSignal], List[str]],
-                 spectrum: DADSpectrum = None,
+                 dad_spectrum: DADSpectrum = None,
+                 ms_spectrum: MSSpectrum = None,
                  directory: str = None,
                  ):
         """
@@ -743,7 +991,7 @@ class HPLCSample(HPLCSampleInfo):
         :param datetimestamp: date and time stamp for when the sample was run
         :param method_name: name of method used to run the sample
         :param signals: list of signals associated with the run
-        :param spectrum: DADSpectrum object with loaded data
+        :param dad_spectrum: DADSpectrum object with loaded data
         :param directory: directory path where the sample may be found
         """
         HPLCSampleInfo.__init__(
@@ -753,7 +1001,8 @@ class HPLCSample(HPLCSampleInfo):
             method_name=method_name,
             signals=signals,
         )
-        self.spectrum: DADSpectrum = spectrum
+        self.spectrum: DADSpectrum = dad_spectrum
+        self.ms_spectrum: MSSpectrum = ms_spectrum
         self.signals = [DADSignal.create_from_DADSignalInfo(signal, self.spectrum) for signal in self.signals]
         self.directory = directory
 
@@ -770,10 +1019,18 @@ class HPLCSample(HPLCSampleInfo):
         if file_path.suffix != '.D' is False or os.path.isdir(file_path) is False:
             raise ValueError(f'The file path "{file_path}" does not appear to be a valid Agilent sample directory')
         sample_info_values = HPLCSampleInfo.get_values_from_acaml(file_path)
-        spectrum = DADSpectrum.create_from_D_file(file_path)
+        try:
+            dad_spectrum = DADSpectrum.create_from_D_file(file_path)
+        except FileNotFoundError:
+            dad_spectrum = None
+        try:
+            ms_spectrum = MSSpectrum.create_from_D_file(file_path)
+        except FileNotFoundError:
+            ms_spectrum = None
         return cls(
             **sample_info_values,
-            spectrum=spectrum,
+            dad_spectrum=dad_spectrum,
+            ms_spectrum=ms_spectrum,
             directory=file_path,
         )
 
